import glob
import os
from typing import Union

import h5py
import numpy as np
from scipy import ndimage as ndi
from scipy.special import softmax
from skimage.measure import regionprops_table
from skimage.segmentation import clear_border

from ovary_analysis.io.hdf5_io import (
    load_im_from_hdf,
    write_multi_dataset_hdf,
)
from ovary_analysis.utils.submission_utils import get_fold_directories


def process_follicle_labels(
    follicle_labels: np.ndarray, volume_threshold: float
) -> np.ndarray:
    """Process follicle labels to remove follicles with volumes
    below the specified threshold. Also removes labels touching
    the image border.

    Parameters
    ----------
    follicle_labels : np.ndarray
        The follicle labels as a label image.
    volume_threshold : float
        The minimum follicle volume in voxels.

    Returns
    -------
    follicle_labels : np.ndarray
        The follicle labels with the small and border-touching
        labels removed.
    """
    follicle_mask_clean = remove_small_follicles(
        follicle_mask=follicle_labels, volume_threshold=volume_threshold
    )

    follicle_mask_no_border = clear_border(follicle_mask_clean)
    follicle_labels, _ = ndi.label(follicle_mask_no_border)

    return follicle_labels


def segment_follicles(
    follicle_prediction: np.ndarray,
    ovary_seg: np.ndarray,
    threshold: float,
    prediction_index: int = 1,
) -> np.ndarray:
    """Segment follicles from a prediction image

    Parameters
    ----------
    follicle_prediction : np.ndarray
        The prediction output from the segmentation model.
        Should be ordered CZYX. Where C is the prediction
        class.
    ovary_seg : np.ndarray
        The ovary segmentation image used to remove follicles detected
        outside of the ovary. Should be a binary image.
    threshold : float
        The probability threshold. Probabilities greater than the threshold
        are called as follicle.
    prediction_index : int
        The index in the prediction image corresponding to the follicle
        prediction. The default value is 1.

    Returns
    -------
    follicle_mask : np.ndarray
        The binary mask of the follicle segmentation.
    """
    follicle_prediction = follicle_prediction[prediction_index, ...]
    follicle_mask = np.zeros_like(follicle_prediction, dtype=bool)

    follicle_mask[follicle_prediction >= threshold] = True
    follicle_mask[np.logical_not(ovary_seg)] = False

    return follicle_mask


def remove_small_follicles(
    follicle_mask: np.ndarray, volume_threshold: float
) -> np.ndarray:
    """Remove follicles below a size threshold

    Parameters
    ----------
    follicle_mask : np.ndarray
        The binary mask of the follicle segmentation.
    volume_threshold : float
        The volume threshold below which follicles are discarded.
        The threshold is in voxels.

    Returns
    -------
    clean_follicles : np.ndarray
        The binary mask of the follicle segmentation with small
        follicles removed.
    """
    label_im, _ = ndi.label(follicle_mask)
    rp_table = regionprops_table(label_im, properties=('label', 'area'))

    area_to_remove_mask = rp_table['area'] < volume_threshold
    labels_to_remove = rp_table['label'][area_to_remove_mask]

    clean_follicles = label_im.copy()
    for label_index in labels_to_remove:
        clean_follicles[clean_follicles == label_index] = 0

    return clean_follicles.astype(bool)


def post_process_follicles(
    follicle_prediction: np.ndarray,
    ovary_seg: np.ndarray,
    prob_threshold: float,
    volume_threshold: float,
    prediction_index: int = 1,
    apply_softmax: bool = True,
) -> np.ndarray:
    """Create a segmentation label image from the model prediction.

    Parameters
    ----------
    follicle_prediction : np.ndarray
        The prediction output from the segmentation model.
        Should be ordered CZYX. Where C is the prediction
        class.
    ovary_seg : np.ndarray
        The ovary segmentation image used to remove follicles detected
        outside of the ovary. Should be a binary image.
    prob_threshold : float
        The probability threshold. Probabilities greater than the threshold
        are called as follicle.
    volume_threshold : float
        The volume threshold below which follicles are discarded.
        The threshold is in voxels.
    prediction_index : int
        The index in the prediction image corresponding to the follicle
        prediction. The default value is 1.
    apply_softmax : bool
        Flag set to true if a softmax should be applied along axis 0
        of the follicle_prediction. The default value is True.

    Returns
    -------
    follicle_labels : np.ndarray
        The follicle segmentation label image.
    """
    if apply_softmax is True:
        follicle_prediction = softmax(follicle_prediction, axis=0)

    follicle_seg = segment_follicles(
        follicle_prediction=follicle_prediction,
        ovary_seg=ovary_seg,
        threshold=prob_threshold,
        prediction_index=prediction_index,
    )

    follicle_mask_clean = remove_small_follicles(
        follicle_mask=follicle_seg, volume_threshold=volume_threshold
    )

    follicle_mask_no_border = clear_border(follicle_mask_clean)

    follicle_labels, _ = ndi.label(follicle_mask_no_border)

    return follicle_labels


def segment_follicles_from_prediction(
    prediction_path: Union[str, os.PathLike],
    raw_im_path: Union[str, os.PathLike],
    prob_threshold: float,
    volume_threshold: float,
    output_dir: Union[str, os.PathLike],
    prediction_index: int = 1,
    apply_softmax: bool = True,
    raw_im_key: str = 'raw_rescaled',
    follicle_labels_key: str = 'follicle_labels_rescaled',
    ovary_segmentation_key: str = 'ovary_labels_rescaled',
    prediction_key: str = 'predictions',
) -> str:
    # load the images
    prediction = load_im_from_hdf(
        fpath=prediction_path, dataset_key=prediction_key
    )

    with h5py.File(raw_im_path, 'r') as f_raw:
        raw_rescaled = f_raw[raw_im_key][:]
        rescaled_attrs = dict(f_raw[raw_im_key].attrs.items())

    try:
        follicle_labels = load_im_from_hdf(
            fpath=raw_im_path, dataset_key=follicle_labels_key
        )
        clean_follicle_lables = process_follicle_labels(
            follicle_labels=follicle_labels, volume_threshold=volume_threshold
        )
    except KeyError:
        follicle_labels = None
        clean_follicle_lables = None
    ovary_segmentation = load_im_from_hdf(
        fpath=raw_im_path, dataset_key=ovary_segmentation_key
    )

    # segment the follicles
    follicles_segmentation = post_process_follicles(
        follicle_prediction=prediction,
        ovary_seg=ovary_segmentation.astype(bool),
        prob_threshold=prob_threshold,
        volume_threshold=volume_threshold,
        prediction_index=prediction_index,
        apply_softmax=apply_softmax,
    )

    # save the segmentation
    output_fname = os.path.basename(prediction_path).replace(
        '_predictions.', '_segmentation.'
    )
    output_fpath = os.path.join(output_dir, output_fname)
    rescaled_dataset = {'data': raw_rescaled, 'attrs': rescaled_attrs}
    writer_kwargs = {
        'file_path': output_fpath,
        'compression': 'gzip',
        'raw_rescaled': rescaled_dataset,
        'follicle_segmentation_rescaled': follicles_segmentation,
        'ovary_segmentation_rescaled': ovary_segmentation,
    }
    if clean_follicle_lables is not None:
        writer_kwargs.update(
            {'follicle_labels_rescaled': clean_follicle_lables}
        )
    write_multi_dataset_hdf(**writer_kwargs)

    return output_fpath


def segment_follicles_fold(
    fold_dir_path: Union[str, os.PathLike],
    raw_im_dir_name: Union[str, os.PathLike],
    output_dir_name: Union[str, os.PathLike],
    predictions_dir_name: str,
    prob_threshold: float,
    volume_threshold: float,
    prediction_index: int = 1,
    apply_softmax: bool = True,
    raw_im_key: str = 'raw_rescaled',
    follicle_labels_key: str = 'follicle_labels_rescaled',
    ovary_segmentation_key: str = 'ovary_labels_rescaled',
    prediction_key: str = 'predictions',
):

    output_dir_fpath = os.path.join(fold_dir_path, output_dir_name)
    if not os.path.isdir(output_dir_fpath):
        os.mkdir(output_dir_fpath)

    # get the prediction directory path
    prediction_dir_path = os.path.join(fold_dir_path, predictions_dir_name)

    # get the raw images
    raw_im_dir = os.path.join(fold_dir_path, raw_im_dir_name)
    raw_im_pattern = os.path.join(raw_im_dir, '*.h5')
    raw_im_files = glob.glob(raw_im_pattern)

    segmentation_fpaths = []
    for raw_im_path in raw_im_files:
        prediction_im_name = os.path.basename(raw_im_path).replace(
            '.h5', '_predictions.h5'
        )
        prediction_path = os.path.join(prediction_dir_path, prediction_im_name)

        seg_path = segment_follicles_from_prediction(
            prediction_path=prediction_path,
            raw_im_path=raw_im_path,
            prob_threshold=prob_threshold,
            volume_threshold=volume_threshold,
            output_dir=output_dir_fpath,
            prediction_index=prediction_index,
            apply_softmax=apply_softmax,
            raw_im_key=raw_im_key,
            follicle_labels_key=follicle_labels_key,
            ovary_segmentation_key=ovary_segmentation_key,
            prediction_key=prediction_key,
        )
        segmentation_fpaths.append(seg_path)

    return segmentation_fpaths


def post_process_follicles_cross_val(
    crossval_dir_path: Union[str, os.PathLike],
    fold_dir_pattern: str,
    raw_im_dir_name: Union[str, os.PathLike],
    predictions_dir_name: str,
    output_dir_name: Union[str, os.PathLike],
    prob_threshold: float,
    volume_threshold: float,
    prediction_index: int = 1,
    apply_softmax: bool = True,
    raw_im_key: str = 'raw_rescaled',
    follicle_labels_key: str = 'follicle_labels_rescaled',
    ovary_segmentation_key: str = 'ovary_labels_rescaled',
    prediction_key: str = 'predictions',
):
    fold_directories = get_fold_directories(
        base_dir=crossval_dir_path, fold_dir_pattern=fold_dir_pattern
    )

    for _, fold_directory in fold_directories:
        _ = segment_follicles_fold(
            fold_dir_path=fold_directory,
            raw_im_dir_name=raw_im_dir_name,
            output_dir_name=output_dir_name,
            predictions_dir_name=predictions_dir_name,
            prob_threshold=prob_threshold,
            volume_threshold=volume_threshold,
            prediction_index=prediction_index,
            apply_softmax=apply_softmax,
            raw_im_key=raw_im_key,
            follicle_labels_key=follicle_labels_key,
            ovary_segmentation_key=ovary_segmentation_key,
            prediction_key=prediction_key,
        )



import datetime
import os
from typing import Optional, Union, Dict, Any

import h5py
import numpy as np
import pandas as pd
from skimage.morphology import binary_dilation, binary_erosion, cube
from skimage.segmentation import find_boundaries

from ..io.hdf5_io import write_training_data

from ..utils.config_utils import (
    _load_and_update_training_config,
    _load_and_update_prediction_config,
)
from ..utils.image import crop_raw_image, crop_label_image

from ..utils.submission_utils import (
    _write_train_script,
    _write_predict_script,
    write_batch_submission_script,
    get_fold_directories,
)
from ..utils.time import date_to_datestring


def create_follicle_weights(
    ovary_im: np.ndarray,
    follicles_im: np.ndarray,
    padding_weight: float = 0,
    ovary_weight: float = 0.8,
    follicle_weight: float = 0.8,
    boundary_weight: float = 1,
    follicle_dilation: int = 4,
    follicle_erosion: int = 2,
):
    ovary_mask = ovary_im.astype(np.bool)
    follicles_mask = follicles_im.astype(np.bool)

    weights = np.zeros_like(ovary_mask, dtype=np.float)

    # apply the padding weights
    weights[np.logical_not(ovary_mask)] = padding_weight

    # apply the ovary weights
    weights[ovary_mask] = ovary_weight

    # apply the follicle weights
    weights[follicles_mask] = follicle_weight

    # apply the follicle boundary weights
    dilated_follicles = binary_dilation(
        follicles_mask, selem=cube(follicle_dilation)
    )
    eroded_follicles = binary_erosion(
        follicles_mask, selem=cube(follicle_erosion)
    )
    boundary_mask = np.logical_xor(dilated_follicles, eroded_follicles)
    weights[boundary_mask] = boundary_weight

    return weights


def create_multiclass_labels(
    follicle_labels: np.ndarray, follicle_border_mode: str = 'outer'
) -> np.ndarray:
    follicles_mask = follicle_labels.astype(bool)
    multiclass_labels = np.zeros_like(follicle_labels, dtype=np.int)
    multiclass_labels[follicles_mask] = 1

    # create periphery region
    follicle_boundaries = find_boundaries(
        follicles_mask, mode=follicle_border_mode
    )

    # set labels for periphery
    multiclass_labels[follicle_boundaries] = 2

    multiclass_one_hot = np.stack(
        [follicles_mask, follicle_boundaries]
    ).astype(np.uint8)

    return multiclass_labels, multiclass_one_hot


def create_follicle_dataset(
    ovary_segmentation_fpath: Union[str, os.PathLike],
    ovary_dataset_fname: str,
    output_directory: str,
    patient_id: str,
    cycle_id: str,
    side: str,
    day: datetime.date,
    mask_raw_im: bool = True,
    min_shape=[80, 80, 80],
    raw_name: str = 'raw_rescaled',
    ovary_seg_name: str = 'ovary_segmentation_rescaled',
    follicle_name: str = 'follicle_labels',
    ovary_name: str = 'ovary_labels',
    weights_name: str = 'weights',
    follicle_border_mode: str = 'outer',
    follicle_dilation: int = 4,
    follicle_erosion: int = 2,
    padding_weight: float = 0,
    ovary_weight: float = 0.8,
    follicle_weight: float = 0.8,
    boundary_weight: float = 1,
    multiclass_name: Optional[str] = None,
):

    # load the ovary segmentation
    with h5py.File(ovary_segmentation_fpath) as f_ovary:
        ovary_segmentation = f_ovary[ovary_seg_name][:]
        raw_im = f_ovary[raw_name][:]
        follicle_labels = f_ovary[follicle_name][:]
        _ = f_ovary[ovary_name][:]

    # make sure the follicle labels are 0 or 1
    follicle_labels[follicle_labels > 0] = 1
    follicle_labels[follicle_labels == 0] = 0

    # crop the data
    ovary_mask = ovary_segmentation > 0
    cropped_raw, _, _ = crop_raw_image(
        raw_im, ovary_mask, min_shape=min_shape, mask_periphery=mask_raw_im
    )
    cropped_follicle, _, _ = crop_label_image(
        follicle_labels, ovary_mask, min_shape=min_shape
    )
    cropped_ovary, _, _ = crop_label_image(
        ovary_mask.copy(), ovary_mask, min_shape=min_shape
    )

    # create the weights image
    weights = create_follicle_weights(
        ovary_im=cropped_ovary,
        follicles_im=cropped_follicle,
        padding_weight=padding_weight,
        ovary_weight=ovary_weight,
        follicle_weight=follicle_weight,
        boundary_weight=boundary_weight,
        follicle_dilation=follicle_dilation,
        follicle_erosion=follicle_erosion,
    )

    # create the filename
    date_string = date_to_datestring(day)
    fname = f'{patient_id}_{cycle_id}_{side}_{date_string}_follicles.h5'
    file_path = os.path.join(output_directory, fname)

    if multiclass_name is not None:
        multiclass_im, multiclass_one_hot = create_multiclass_labels(
            cropped_follicle, follicle_border_mode=follicle_border_mode
        )

    else:
        multiclass_im = None
        multiclass_one_hot = None
    # write the data
    write_training_data(
        raw_im=cropped_raw,
        raw_im_path=ovary_dataset_fname,
        follicle_im=cropped_follicle,
        follicle_im_path=ovary_dataset_fname,
        ovary_im=cropped_ovary,
        ovary_im_path=ovary_segmentation_fpath,
        file_path=file_path,
        raw_name=raw_name,
        follicle_name=follicle_name,
        ovary_name=ovary_name,
        weights_im=weights,
        weights_name=weights_name,
        multiclass_im=multiclass_im,
        multiclass_name=multiclass_name,
        multiclass_one_hot_im=multiclass_one_hot,
        compression='gzip',
    )

    return file_path


def batch_create_follicle_dataset(
    dataset_table: pd.DataFrame,
    ovary_datasets_dir: str,
    ovary_segmentations_dir: str,
    output_base_dir: str,
    raw_name: str = 'raw',
    follicle_name: str = 'follicle_labels',
    ovary_name: str = 'ovary_labels',
    ovary_seg_name: str = 'ovary_segmentation_rescaled',
    weights_name: str = 'weights',
    mask_raw_im: bool = True,
    min_shape=[80, 80, 80],
    follicle_border_mode='outer',
    follicle_dilation: int = 4,
    follicle_erosion: int = 2,
    padding_weight: float = 0,
    ovary_weight: float = 0.8,
    follicle_weight: float = 0.8,
    boundary_weight: float = 1,
    multiclass_name: Optional[str] = None,
):

    # create the output directories
    group_names = dataset_table['group'].unique()
    for group in group_names:
        group_path = os.path.join(output_base_dir, group)
        if not os.path.isdir(group_path):
            os.mkdir(group_path)

    file_paths = []

    for _, row in dataset_table.iterrows():
        patient_id = row['patient']
        cycle_id = row['cycle']
        side = row['side']
        day = row['day']
        group = row['group']

        if isinstance(day, str):
            date_time_obj = datetime.datetime.strptime(day, '%Y-%m-%d')
            date_string = date_time_obj.strftime('%y%m%d')
        else:
            date_string = day.strftime('%y%m%d')
        filename_base = f'{patient_id}_{cycle_id}_{side}_{date_string}_ovary'

        ovary_segmentation_fpath = os.path.join(
            ovary_segmentations_dir, filename_base + '_segmentation.h5'
        )
        ovary_dataset_fpath = os.path.join(
            ovary_datasets_dir, filename_base + '.h5'
        )
        output_directory = os.path.join(output_base_dir, group)

        file_path = create_follicle_dataset(
            ovary_segmentation_fpath=ovary_segmentation_fpath,
            ovary_dataset_fname=ovary_dataset_fpath,
            output_directory=output_directory,
            patient_id=patient_id,
            cycle_id=cycle_id,
            side=side,
            day=day,
            mask_raw_im=mask_raw_im,
            min_shape=min_shape,
            raw_name=raw_name,
            follicle_name=follicle_name,
            ovary_name=ovary_name,
            ovary_seg_name=ovary_seg_name,
            weights_name=weights_name,
            follicle_border_mode=follicle_border_mode,
            follicle_dilation=follicle_dilation,
            follicle_erosion=follicle_erosion,
            padding_weight=padding_weight,
            ovary_weight=ovary_weight,
            follicle_weight=follicle_weight,
            boundary_weight=boundary_weight,
            multiclass_name=multiclass_name,
        )

        file_paths.append(file_path)

    dataset_table['follicle_path'] = file_paths

    return dataset_table


def create_follicle_datasets_from_fold(
    data_table_name: str,
    fold_dir: Union[str, os.PathLike],
    ovary_seg_dir_name: str,
    ovary_datasets_dir: Union[str, os.PathLike],
    output_base_dir: str,
    raw_name: str = 'raw',
    follicle_name: str = 'follicle_labels',
    ovary_name: str = 'ovary_labels',
    ovary_seg_name: str = 'ovary_segmentation_rescaled',
    weights_name: str = 'weights',
    mask_raw_im: bool = True,
    min_shape=[80, 80, 80],
    follicle_border_mode: str = 'outer',
    follicle_dilation: int = 4,
    follicle_erosion: int = 2,
    padding_weight: float = 0,
    ovary_weight: float = 0.8,
    follicle_weight: float = 0.8,
    boundary_weight: float = 1,
    multiclass_name: Optional[str] = None,
) -> pd.DataFrame:
    # load the dataset table
    dataset_table_fpath = os.path.join(fold_dir, data_table_name)
    dataset_table = pd.read_csv(dataset_table_fpath)

    # get the name of the ovary segmentations directory
    ovary_segmentations_dir = os.path.join(fold_dir, ovary_seg_dir_name)

    # create the follicle datasets
    updated_dataset_table = batch_create_follicle_dataset(
        dataset_table=dataset_table,
        ovary_datasets_dir=ovary_datasets_dir,
        ovary_segmentations_dir=ovary_segmentations_dir,
        output_base_dir=output_base_dir,
        raw_name=raw_name,
        follicle_name=follicle_name,
        ovary_name=ovary_name,
        ovary_seg_name=ovary_seg_name,
        weights_name=weights_name,
        mask_raw_im=mask_raw_im,
        min_shape=min_shape,
        follicle_border_mode=follicle_border_mode,
        follicle_dilation=follicle_dilation,
        follicle_erosion=follicle_erosion,
        padding_weight=padding_weight,
        ovary_weight=ovary_weight,
        follicle_weight=follicle_weight,
        boundary_weight=boundary_weight,
        multiclass_name=multiclass_name,
    )

    return updated_dataset_table


def create_cross_validation_datasets(
    crossval_dir: Union[str, os.PathLike],
    ovary_data_table_name: str,
    follicle_data_table_name: str,
    ovary_seg_dir_name: str,
    ovary_datasets_dir: Union[str, os.PathLike],
    training_config_path: Union[str, os.PathLike],
    training_runner: str,
    training_job_settings: Dict[str, Any],
    prediction_config_path: Union[str, os.PathLike],
    prediction_runner: str,
    prediction_job_settings: Dict[str, Any],
    output_base_dir: str,
    raw_name: str = 'raw',
    follicle_name: str = 'follicle_labels',
    ovary_name: str = 'ovary_labels',
    ovary_seg_name: str = 'ovary_segmentation_rescaled',
    weights_name: str = 'weights',
    mask_raw_im: bool = True,
    min_shape=[80, 80, 80],
    follicle_border_mode: str = 'outer',
    follicle_dilation: int = 4,
    follicle_erosion: int = 2,
    padding_weight: float = 0,
    ovary_weight: float = 0.8,
    follicle_weight: float = 0.8,
    boundary_weight: float = 1,
    multiclass_name: Optional[str] = None,
):
    # get the fold directories
    fold_dirs = get_fold_directories(crossval_dir, 'fold_*')
    n_fold = len(fold_dirs)
    print(f'processing {n_fold} folds')

    training_scripts = []
    prediction_scripts = []
    if not os.path.isdir(output_base_dir):
        os.mkdir(output_base_dir)
    for fold_index, fold in fold_dirs:
        fold_name = os.path.basename(fold)
        fold_dir_path = os.path.join(output_base_dir, fold_name)

        print(fold_dir_path)
        if not os.path.isdir(fold_dir_path):
            os.mkdir(fold_dir_path)

        # create the follicle datasets
        data_table = create_follicle_datasets_from_fold(
            data_table_name=ovary_data_table_name,
            fold_dir=fold,
            ovary_seg_dir_name=ovary_seg_dir_name,
            ovary_datasets_dir=ovary_datasets_dir,
            output_base_dir=fold_dir_path,
            raw_name=raw_name,
            follicle_name=follicle_name,
            ovary_name=ovary_name,
            ovary_seg_name=ovary_seg_name,
            weights_name=weights_name,
            mask_raw_im=mask_raw_im,
            min_shape=min_shape,
            follicle_border_mode=follicle_border_mode,
            follicle_dilation=follicle_dilation,
            follicle_erosion=follicle_erosion,
            padding_weight=padding_weight,
            ovary_weight=ovary_weight,
            follicle_weight=follicle_weight,
            boundary_weight=boundary_weight,
            multiclass_name=multiclass_name,
        )

        # save the datatable
        table_output_fpath = os.path.join(
            fold_dir_path, follicle_data_table_name
        )
        data_table.to_csv(table_output_fpath)

        # setup and save the training files
        checkpoint_dir = os.path.join(fold_dir_path, 'checkpoints')
        train_data_path = os.path.join(fold_dir_path, 'train')
        val_data_path = os.path.join(fold_dir_path, 'val')
        updated_training_config_path = _load_and_update_training_config(
            config_path=training_config_path,
            fold_index=fold_index,
            checkpoint_dir=checkpoint_dir,
            train_file_dir=[train_data_path],
            val_file_dir=[val_data_path],
            config_output_directory=fold_dir_path,
        )
        training_script_path = _write_train_script(
            train_runner=training_runner,
            config_path=updated_training_config_path,
            output_dir=fold_dir_path,
        )
        training_scripts.append(training_script_path)

        # setup and save the prediction config
        test_data_path = os.path.join(fold_dir_path, 'test')
        model_path = os.path.join(checkpoint_dir, 'best_checkpoint.pytorch')
        prediction_output_path = os.path.join(fold_dir_path, 'predictions')
        if not os.path.isdir(prediction_output_path):
            os.mkdir(prediction_output_path)
        updated_prediction_config_path = _load_and_update_prediction_config(
            config_path=prediction_config_path,
            fold_index=fold_index,
            model_path=model_path,
            image_files=[train_data_path, val_data_path, test_data_path],
            prediction_output_directory=prediction_output_path,
            config_output_directory=fold_dir_path,
        )
        prediction_script_path = _write_predict_script(
            predict_runner=prediction_runner,
            config_path=updated_prediction_config_path,
            output_dir=fold_dir_path,
        )
        prediction_scripts.append(prediction_script_path)

        print(f'fold{fold_index} complete')

    print('writing submission scripts')
    # write the training submission scripts
    training_submission_path = os.path.join(
        output_base_dir, 'training_submission_script.sh'
    )
    write_batch_submission_script(
        submission_script_path=training_submission_path,
        runner_scripts=training_scripts,
        job_prefix='training',
        job_settings=training_job_settings,
    )

    # write the prediction submission scripts
    prediction_submission_path = os.path.join(
        output_base_dir, 'prediction_submission_script.sh'
    )
    write_batch_submission_script(
        submission_script_path=prediction_submission_path,
        runner_scripts=prediction_scripts,
        job_prefix='prediction',
        job_settings=prediction_job_settings,
    )

    print('done')



import os
from typing import Any, Dict, Optional, Tuple, Union

import h5py
import numpy as np


def load_raw_im(
    fpath: str, raw_name: str = 'raw', rescaled_name: str = 'raw_rescaled'
) -> Tuple[np.ndarray, Dict[str, Any], np.ndarray, Dict[str, Any]]:

    with h5py.File(fpath, 'r') as f_raw:
        raw_im = f_raw[raw_name][:]
        raw_attrs = dict(f_raw[raw_name].attrs.items())
        rescaled_im = f_raw[rescaled_name][:]
        rescaled_attrs = dict(f_raw[rescaled_name].attrs.items())

    return raw_im, raw_attrs, rescaled_im, rescaled_attrs


def load_im_from_hdf(
    fpath: Union[str, os.PathLike], dataset_key: str
) -> np.ndarray:
    """Load an image from an hdf5 file.

    Parameters
    ----------
    fpath : Union[str, os.PathLike]
        The path to the image file.
    dataset_key : str
        The key to load the data from.

    Returns
    -------
    im : np.ndarray
        The image that was loaded.
    """
    with h5py.File(fpath, 'r') as f:
        im = f[dataset_key][:]
    return im


def write_hdf(
    pixel_array: np.ndarray,
    fname: str,
    attrs: Optional[Dict[str, Any]] = None,
    dataset_name: str = 'image',
    compression: Optional[str] = None,
):
    """Write a pixel array to an hdf5 file

    Parameters
    ----------
    pixel_array : np.ndarray
        The pixel intensity data to write.
    fname : str
        The name of the file to write
    dataset_name : str
        The name for the dataset containing the pixel array.
        The default value is 'image'
    compression : Optional[str]
        The compression filter to apply to the data. Can be "gzip" or None.
        The default value is None.
    """
    with h5py.File(fname, 'w') as f:
        dset = f.create_dataset(
            dataset_name,
            pixel_array.shape,
            dtype=pixel_array.dtype,
            data=pixel_array,
            compression=compression,
        )
        if attrs is not None:
            for k, v in attrs.items():
                dset[k] = v


def write_raw(
    raw_im: np.ndarray,
    raw_px_size: Tuple[float, float, float],
    raw_path: str,
    rescaled_im: np.ndarray,
    rescaled_px_size: Tuple[float, float, float],
    fname: str,
    raw_name: str = 'raw',
    rescaled_name: str = 'raw_rescaled',
    compression: Optional[str] = None,
):
    """Write a pixel array to an hdf5 file

    Parameters
    ----------
    raw_im : np.ndarray
        The raw pixel intensity data to write.
    raw_px_size : Tuple[float, float, float]
        The size of each voxel in the raw image in mm / voxel
    raw_path : str
        The file path to the original raw image
    rescaled_im : np.ndarray
        The rescaled raw pixel intensity data to write.
    rescaled_px_size : Tuple[float, float, float]
        The size of each voxel in the rescaled image in mm / voxel
    fname : str
        The name of the file to write
    raw_name : str
        The name for the dataset containing the raw pixel array.
        The default value is 'raw'
    rescaled_name : str
        The name for the dataset containing the rescaled pixel array.
        The default value is 'raw_rescaled'
    compression : Optional[str]
        The compression filter to apply to the data. Can be "gzip" or None.
        The default value is None.
    """
    with h5py.File(fname, 'w') as f:
        raw = f.create_dataset(
            raw_name,
            raw_im.shape,
            dtype=raw_im.dtype,
            data=raw_im,
            compression=compression,
        )
        raw.attrs['px_size'] = raw_px_size
        raw.attrs['orig_path'] = raw_path

        rescaled = f.create_dataset(
            rescaled_name,
            rescaled_im.shape,
            dtype=rescaled_im.dtype,
            data=rescaled_im,
            compression=compression,
        )
        rescaled.attrs['px_size'] = rescaled_px_size


def write_training_data(
    raw_im: np.ndarray,
    follicle_im: np.ndarray,
    ovary_im: np.ndarray,
    file_path: str,
    raw_name: str = 'raw',
    raw_im_path: str = '',
    follicle_name: str = 'follicle_labels',
    follicle_im_path: str = '',
    ovary_name: str = 'ovary_labels',
    ovary_im_path: str = '',
    weights_im: Optional[np.ndarray] = None,
    weights_name: str = 'weights',
    multiclass_im: Optional[np.ndarray] = None,
    multiclass_name: str = 'multiclass_name',
    multiclass_one_hot_im: Optional[np.ndarray] = None,
    multiclass_one_hot_name: str = 'multiclass_one_hot',
    compression: Optional[str] = None,
):
    """Write a training dataset to an hdf5 file

    Parameters
    ----------
    raw_im : np.ndarray
        The raw image pixel intensity data to write.
    label_im : np.ndarray
        The the labeled image data to write.
    file_path : str
        The name of the file to write
    raw_name : str
        The name for the dataset containing the raw pixel array.
        The default value is 'raw'
    raw_name : str
        The name for the dataset containing the label pixel array.
        The default value is 'labels'
    compression : Optional[str]
        The compression filter to apply to the data. Can be "gzip" or None.
        The default value is None.
    """
    with h5py.File(file_path, 'w') as f:
        # write the image
        raw = f.create_dataset(
            raw_name,
            raw_im.shape,
            dtype=raw_im.dtype,
            data=raw_im,
            compression=compression,
        )
        raw.attrs['path'] = raw_im_path
        # write the labels
        follicles = f.create_dataset(
            follicle_name,
            follicle_im.shape,
            dtype=follicle_im.dtype,
            data=follicle_im,
            compression=compression,
        )
        follicles.attrs['path'] = follicle_im_path
        ovary = f.create_dataset(
            ovary_name,
            ovary_im.shape,
            dtype=ovary_im.dtype,
            data=ovary_im,
            compression=compression,
        )
        ovary.attrs['path'] = ovary_im_path

        if weights_im is not None:
            f.create_dataset(
                weights_name,
                weights_im.shape,
                dtype=weights_im.dtype,
                data=weights_im,
                compression=compression,
            )

        if multiclass_im is not None:
            f.create_dataset(
                multiclass_name,
                multiclass_im.shape,
                dtype=multiclass_im.dtype,
                data=multiclass_im,
                compression=compression,
            )
        if multiclass_one_hot_im is not None:
            f.create_dataset(
                multiclass_one_hot_name,
                multiclass_one_hot_im.shape,
                dtype=multiclass_one_hot_im.dtype,
                data=multiclass_one_hot_im,
                compression=compression,
            )


def _write_dataset_from_array(
    file_handle: h5py.File,
    dataset_name: str,
    dataset_array: np.ndarray,
    compression: str = 'gzip',
):
    file_handle.create_dataset(
        dataset_name,
        dataset_array.shape,
        dtype=dataset_array.dtype,
        data=dataset_array,
        compression=compression,
    )


def _write_dataset_from_dict(
    file_handle: h5py.File,
    dataset_name: str,
    dataset: Dict[str, Any],
    compression: str = 'gzip',
):
    dataset_array = dataset['data']
    dset = file_handle.create_dataset(
        dataset_name,
        dataset_array.shape,
        dtype=dataset_array.dtype,
        data=dataset_array,
        compression=compression,
    )

    dataset_attrs = dataset.get('attrs', None)
    if dataset_attrs is not None:
        for k, v in dataset_attrs.items():
            dset.attrs[k] = v


def write_multi_dataset_hdf(
    file_path: str, compression: str = 'gzip', **kwargs
):
    if len(kwargs) == 0:
        raise ValueError('Must supply at least one dataset as a kwarg')
    with h5py.File(file_path, 'w') as f:
        for k, v in kwargs.items():
            if isinstance(v, np.ndarray):
                _write_dataset_from_array(
                    file_handle=f,
                    dataset_name=k,
                    dataset_array=v,
                    compression=compression,
                )
            else:
                _write_dataset_from_dict(
                    file_handle=f,
                    dataset_name=k,
                    dataset=v,
                    compression=compression,
                )



from ovary_analysis.post_process.ovary import post_process_ovary_cross_val

CROSS_VAL_DIR = '/cluster/work/cobi/kevin/ivf/cross_validation_20211031/ovary_crossval'
OVARY_DATASETS_DIR = '/cluster/work/cobi/kevin/ivf/cross_validation_20211031/converted_data/ovary_datasets_20211101'
PROBABILITY_THRESHOLD = 0.8
APPLY_SIGMOID = True
DILATION_SIZE = 5

post_process_ovary_cross_val(
    crossval_dir=CROSS_VAL_DIR,
    ovary_datasets_dir=OVARY_DATASETS_DIR,
    predictions_subdir='predictions',
    output_dir_name='segmentations',
    threshold=PROBABILITY_THRESHOLD,
    apply_sigmoid=APPLY_SIGMOID,
    prediction_index=0,
    dilation_size=DILATION_SIZE,
    raw_im_key='raw_rescaled',
    follicle_labels_key='follicle_labels_rescaled',
    ovary_labels_key='ovary_labels_rescaled',
    prediction_key='predictions',
)



from typing import Any, Dict, List, Union
import os
import yaml


def _load_config(config_path: Union[str, os.PathLike]) -> Dict[str, Any]:
    """Load a YAML configuration file.

    Parameters
    ----------
    config_path : Union[str, os.PathLike]
        The path to the configuration file.

    Returns
    -------
    config : Dict[str, Any]
        The configuration file loaded into a dictionary.
    """
    with open(config_path) as f_config:
        config = yaml.safe_load(f_config)
    return config


def _load_and_update_training_config(
    config_path: Union[str, os.PathLike],
    fold_index: int,
    checkpoint_dir: Union[str, os.PathLike],
    train_file_dir: List[Union[str, os.PathLike]],
    val_file_dir: List[Union[str, os.PathLike]],
    config_output_directory: Union[str, os.PathLike],
) -> str:
    """Load a template training config and update it with the file paths for
    a given experiment.

    The config should be in the format of the pytorch-3dunet library
    https://github.com/wolny/pytorch-3dunet

    Parameters
    ----------
    config_path : Union[str, os.PathLike]
        The path to the template config file.
    fold_index : int
        The index of the fold the prediction config is being updated for.
    checkpoint_dir: Union[str, os.PathLike]
        The directory in which to save the checkpoint files.
    train_file_dir: List[Union[str, os.PathLike]]
        The directory containing the data for training.
    val_file_dir: List[Union[str, os.PathLike]]
        The directory the containing the data for vaildation.
    config_output_directory: Union[str, os.PathLike]
        The directory in which to save the updated config file.

    Returns
    -------
    updated_config_path : str
        The path to which the updated config file was saved.
    """
    config = _load_config(config_path)

    # update checkpoint_dir
    if not os.path.isdir(checkpoint_dir):
        os.mkdir(checkpoint_dir)
    config['trainer']['checkpoint_dir'] = checkpoint_dir

    # update train file paths
    for directory in train_file_dir:
        assert os.path.isdir(directory)
    config['loaders']['train']['file_paths'] = train_file_dir

    # update validation file paths
    for directory in val_file_dir:
        assert os.path.isdir(directory)
    config['loaders']['val']['file_paths'] = val_file_dir

    # save the updated file
    config_name = os.path.basename(config_path).replace(
        '.yaml', f'_fold_{fold_index}.yaml'
    )
    updated_config_path = os.path.join(config_output_directory, config_name)
    with open(updated_config_path, 'w') as outfile:
        yaml.dump(config, outfile, default_flow_style=False)

    return updated_config_path


def _load_and_update_prediction_config(
    config_path: Union[str, os.PathLike],
    fold_index: int,
    model_path: Union[str, os.PathLike],
    image_files: List[Union[str, os.PathLike]],
    prediction_output_directory: Union[str, os.PathLike],
    config_output_directory: Union[str, os.PathLike],
) -> str:
    """Load a template prediction config and update it with the model and file paths for
    a given experiment.

    The config should be in the format of the pytorch-3dunet library
    https://github.com/wolny/pytorch-3dunet

    Parameters
    ----------
    config_path : Union[str, os.PathLike]
        The path to the template config file.
    fold_index : int
        The index of the fold the prediction config is being updated for.
    model_path : Union[str, os.PathLike]
        The path to the model to use for prediction.
    image_files : List[Union[str, os.PathLike]]
        The list of folders to make predictions on.
    prediction_output_directory : Union[str, os.PathLike]
        The directory the predictions will be saved to.
    config_output_directory: Union[str, os.PathLike]
        The directory in which to save the updated config file.

    Returns
    -------
    updated_config_path : str
        The path to which the updated config file was saved.
    """
    config = _load_config(config_path)

    # update model path
    if model_path is not None:
        config['model_path'] = model_path

    # update the prediction output directory
    config['loaders']['output_dir'] = prediction_output_directory

    # update image file paths
    config['loaders']['test']['file_paths'] = image_files

    # save the updated file
    config_name = os.path.basename(config_path).replace(
        '.yaml', f'_fold_{fold_index}.yaml'
    )
    updated_config_path = os.path.join(config_output_directory, config_name)
    with open(updated_config_path, 'w') as outfile:
        yaml.dump(config, outfile, default_flow_style=False)

    return updated_config_path



from ovary_analysis.train.ovary_seg import create_cross_validation

training_job_settings = {
    'n_cpus': 4,
    'mem_per_cpu': 6144,
    'n_gpus': 1,
    'job_time': [96, 0]
}

prediction_job_settings = {
    'n_cpus': 6,
    'mem_per_cpu': 6144,
    'n_gpus': 8,
    'job_time': [0, 30]
}

training_runner = '/cluster/home/kyamauch/.local/lib/python3.8/site-packages/pytorch3dunet/train.py'
prediction_runner = '/cluster/home/kyamauch/.local/lib/python3.8/site-packages/pytorch3dunet/predict.py'

create_cross_validation(
    data_table_path='/cluster/work/cobi/kevin/ivf/cross_validation_20211031/converted_data/ovary_datasets_20211101.csv',
    dataset_dir='/cluster/work/cobi/kevin/ivf/cross_validation_20211031/converted_data/ovary_datasets_20211101',
    training_config_path='/cluster/work/cobi/kevin/ivf/cross_validation_20211031/cross_validation_ovary_train_20211102.yaml',
    training_runner=training_runner,
    training_job_settings=training_job_settings,
    prediction_config_path='/cluster/work/cobi/kevin/ivf/cross_validation_20211031/crossvalidation_ovary_test_20211102.yaml',
    prediction_runner=prediction_runner,
    prediction_job_settings=prediction_job_settings,
    output_dir='/cluster/work/cobi/kevin/ivf/cross_validation_20211031/ovary_crossval',
    df_output_fname='ovary_crossvalidation_20211101.csv',
    n_folds=10,
    frac_train=0.89,
    random_seed=0
)


import multiprocessing as mp
import glob
import os
from functools import partial

import h5py
import numpy as np
import pandas as pd
from scipy import ndimage as ndi
from skimage.measure import regionprops_table

from .measure_utils import binary_mask_to_surface
from ..utils.file_names import parse_seg_fname
from ..utils.submission_utils import get_fold_directories


def measure_surface_area_from_labels(
    label_image: np.ndarray, n_mesh_smoothing_interations: int = 10
) -> pd.DataFrame:
    """Measure the surface measurement of each object in a label image.

    Parameters
    ----------
    label_image : LabelImage
        The label image from which to measure the surface properties.
    n_mesh_smoothing_interations : int
        The number of interations of smooting to perform. Smoothing is
        done by the trimesh laplacian filter:
        https://trimsh.org/trimesh.smoothing.html#trimesh.smoothing.filter_laplacian
        Default value is 50.

    Returns
    -------
    surface_properties_table : LabelMeasurementTable
        The measured surface properties. Each measurement is a column and each
        object is a row.
    """
    if label_image.ndim != 3:
        raise ValueError(
            "measure_surface_properties_from_labels requires a 3D label image"
        )

    all_label_indices = np.unique(label_image)
    surface_areas = []
    labels = []
    for label_index in all_label_indices:
        if label_index == 0:
            # background is assumed to be label 0
            continue
        try:
            label_mask = label_image == label_index
            smoothed_mesh = binary_mask_to_surface(
                label_mask,
                n_mesh_smoothing_interations=n_mesh_smoothing_interations,
            )
            surface_area = smoothed_mesh.area
        except ValueError:
            # if the meshing fails set surface area to 0
            surface_area = 0
        labels.append(label_index)
        surface_areas.append(surface_area)

    surface_properties_table = pd.DataFrame(
        {"label_index": labels, "surface_area": surface_areas}
    )

    return surface_properties_table.set_index("label_index")


def measure_image(
    label_im: np.ndarray, n_mesh_smoothing_interations: int = 10
) -> pd.DataFrame:
    """The volume of the labeled objects.

    Parameters
    ----------
    label_im : str
        directory of labeled three dimensional image
    n_mesh_smoothing_interations : int
        The number of interations of smooting to perform. Smoothing is
        done by the trimesh laplacian filter:
        https://trimsh.org/trimesh.smoothing.html#trimesh.smoothing.filter_laplacian
        Default value is 50.

    Returns
    -------
    df : pandas dataframe
        label of the follicles
        centroid in three dimensional image
        number of pixels included in each labeled objects
    """
    # measure volumes
    props = regionprops_table(
        label_im, properties=("label", "centroid", "area")
    )
    volume_df = pd.DataFrame(props).rename(
        columns={"label": "label_index", "area": "volume"}
    )
    volume_df.set_index("label_index", inplace=True)

    # measure surface area
    surface_area_df = measure_surface_area_from_labels(
        label_image=label_im,
        n_mesh_smoothing_interations=n_mesh_smoothing_interations,
    )

    return pd.concat((volume_df, surface_area_df), axis=1).reset_index()


def process_image(
    file_path: str, follicles_key: str = 'follicle_segmentation_rescaled'
) -> pd.DataFrame:
    """This function takes a file path, loads the image and passes it to the measurement function(s).
    It then returns the dataframe with the measured properties.

    Parameters
    ----------
    file_path : str
        directory of images

    Returns
    -------
    df : pandas dataframe
        dataframe includes filename, label, centroid, and volume from one image
    """
    file_name = os.path.split(file_path)[1]
    file_name = file_name.replace(".h5", "")

    with h5py.File(file_path, "r") as f_raw:
        follicle_labels = f_raw[follicles_key][:]
        label_im, _ = ndi.label(follicle_labels)
        df = measure_image(label_im)
        patient_id, cycle, side, date = parse_seg_fname(file_name)
        df["patient_id"] = patient_id
        df["cycle"] = cycle
        df["side"] = side
        df["date"] = date
        df["file_path"] = file_path
        return df


def measure_dir(
    dir_path: str,
    file_pattern: str = "*.h5",
    follicles_key: str = 'follicle_segmentation_rescaled',
) -> pd.DataFrame:
    """measure the volume and centroid (multi-processing) of all images in one path. concat them into one data frame

    Parameters
    ----------
    dir_path : str
        directory of folder with images
    file_pattern : str
        data type of images

    Returns
    -------
    df : pandas dataframe
        dataframe includes filename, label, centroid, and volume from all images
    """

    # walk through the path and get the directory
    datapath = os.path.join(dir_path, file_pattern)

    # glob code to get the list of files
    image_files = glob.glob(datapath)

    process_func = partial(process_image, follicles_key=follicles_key)

    # multiprocessing to map process_image() to the files
    with mp.get_context("spawn").Pool() as pool:
        results = pool.map(process_func, image_files)

    # concatenate the results into a single df
    df = pd.concat(results, ignore_index=True)
    return df


def measure_cross_val(
    cross_val_dir_path: str,
    segmentation_folder_name: str,
    fold_dir_pattern: str = "fold_*",
    file_pattern: str = "*.h5",
    follicles_key: str = 'follicle_segmentation_rescaled',
) -> pd.DataFrame:
    fold_directories = get_fold_directories(
        base_dir=cross_val_dir_path, fold_dir_pattern=fold_dir_pattern
    )
    measurements = []

    for fold_index, fold_directory in fold_directories:
        segmentation_directory_path = os.path.join(
            fold_directory, segmentation_folder_name
        )
        measurement_table = measure_dir(
            dir_path=segmentation_directory_path,
            file_pattern=file_pattern,
            follicles_key=follicles_key,
        )
        measurement_table['fold_index'] = fold_index
        measurement_table['fold_directory'] = fold_directory
        measurements.append(measurement_table)

    return pd.concat(measurements, ignore_index=True)
